from django.contrib.gis.db import models
from django.utils.translation import gettext_lazy as _
from django.urls import reverse
from django.utils.functional import cached_property

class Station(models.Model):

    wigos_id = models.CharField(_("Wigos ID"), max_length=50, primary_key=True)
    name = models.CharField(_("Station name"), max_length=255)
    geom = models.PointField(_("Geometry"),)
    in_oscar = models.BooleanField(_("In Oscar?"),)

    class Meta:
        verbose_name = _("Station")
        verbose_name_plural = _("Stations")

    def __str__(self):
        return f'{self.name}-{self.wigos_id}'
    
    @cached_property
    def stations_api(self):
        return reverse("stations_api")

        
# Create your models here.
class Transmission(models.Model):

    station = models.ForeignKey("Station", on_delete=models.CASCADE)
    variable = models.CharField(_("Transmission Variable"), max_length=50 )
    received = models.IntegerField(_("Transmissions received"), null=True)
    expected = models.IntegerField(_("Transmissions expected"), null=True)
    received_rate = models.DecimalField(_("Transmission Rate"), max_digits=5, decimal_places=2)
    received_date = models.DateTimeField(_("Date Time Received"), auto_now=False, auto_now_add=False)

    class Meta:
        verbose_name = _("Transmission")
        verbose_name_plural = _("Transmissions")

    def __str__(self):
        return f'{self.station} - {self.variable} - {self.received_date}'
    





from django.shortcuts import render
import pytz
import json

from rest_framework.generics import ListAPIView
from climweb_wdqms.models import Transmission, Station
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework.permissions import BasePermission, IsAuthenticated, SAFE_METHODS
from climweb_wdqms.serializers import StationSerializer
from datetime import datetime
from rest_framework.views import APIView
from rest_framework.response import Response
from django.db.models.functions import ExtractHour,ExtractMonth,ExtractYear
from django.db.models import Avg, Func, F
from django.db.models.functions import TruncMonth
from django.utils.timezone import utc




def validate_params(query_params, supported_params):
    # Check for unsupported parameters
    error_message = None
    unsupported_params = [param for param in query_params.keys() if param not in supported_params]
    
    # If there are unsupported parameters, return an error response
    if unsupported_params:
        error_message = {'error': f'Unsupported parameter(s): {", ".join(unsupported_params)}. Only Supports {", ".join(supported_params)}'}
        # return Response(error_message, status=400)  # Return a 400 Bad Request response
    

    # Check if the 'param' parameter is present in the request
    for q in supported_params:
        if q not in query_params and q != 'station':
            error_message = {'error': f'Parameter "{q}" is required.'}
            # return Response(error_message, status=400)  # Return a 400 Bad Request response


    return error_message

class ReadOnly(BasePermission):
    def has_permission(self, request, view):
        return request.method in SAFE_METHODS
    

class StationListView(ListAPIView):
    queryset = Station.objects.all()
    serializer_class = StationSerializer
    permission_classes = [IsAuthenticated | ReadOnly]

    def get_queryset(self):
        queryset = super().get_queryset()
        wigos_id = self.request.query_params.get('wigos_id')
        if wigos_id:
            queryset = queryset.filter(wigos_id=wigos_id)
        return queryset
    
# Create your views here.
class SynopTransmissionView(APIView):
    filter_backends = [DjangoFilterBackend]
    filterset_fields = ["received_date", "station", "variable"]
    permission_classes = [IsAuthenticated | ReadOnly]

    def get(self, request):
        
        queryset = Transmission.objects.all()
        supported_params = ['station', 'frequency', 'received_date', 'variable']
        query_params = request.query_params
        
        validate = validate_params(query_params, supported_params)

        if validate:
            return Response(validate, status=400)  # Return a 400 Bad Request response

        # query parameters 
        station = request.query_params.get('station', None)
        frequency = request.query_params.get('frequency', None)
        received_date = request.query_params.get('received_date', queryset.order_by('received_date').values_list('received_date').last()[0].strftime("%Y-%m-%dT%H:%M:%SZ"))
        variable = request.query_params.get('variable', 'pressure')

        result = []

        if station is not None:
            queryset = queryset.filter(station = station)

        if received_date:
            received_date = datetime.strptime(f"{received_date}T00:00:00Z", "%Y-%m-%dT%H:%M:%SZ").replace(tzinfo=pytz.UTC)

            # check the frequencies 
            if frequency == 'monthly_synop':
                queryset = queryset.filter(received_date__year = received_date.year, received_date__month = received_date.month, variable=variable).order_by('received_date__hour')
            
            elif frequency == 'daily_synop':
                queryset = queryset.filter(received_date__date=received_date.date(), variable=variable).order_by('received_date__hour')
            elif frequency == 'yearly_synop':
                queryset = queryset.filter(received_date__year = received_date.year, variable=variable).order_by('received_date__hour')

        # Extract the hour from the received_date and annotate the queryset
        queryset = queryset.annotate(
            synop_hour=ExtractHour('received_date', tzinfo=utc)
        )

        # Aggregate the queryset to calculate the average received_rate for each synoptic hour
        queryset = queryset.values('synop_hour').annotate(
            avg_received_rate=Avg('received_rate'),
            avg_received=Avg('received'),
            avg_expected=Avg('expected'),
        )

        # Format the result
        result = [
            {
                'synop_hour': str(hour).zfill(2),  # Format hour to have leading zero if needed
                'avg_received_rate': round(avg_rate, 0), # Round to 0 decimal places
                'avg_received': round(avg_received, 0),  # Round to 0 decimal places
                'avg_expected': round(avg_expected, 0)  # Round to 0 decimal places
            }
            for hour, avg_rate, avg_received, avg_expected in queryset.values_list('synop_hour', 'avg_received_rate','avg_received', 'avg_expected' )
        ] 
            
        # Return response
        return Response(result)


class MonthlyTransmissionView(APIView):

    def get(self, request):

        queryset = Transmission.objects.all()
        result = []

        latest_year =  queryset.values_list('received_date__year').order_by('received_date__year').last()
        supported_params = ['station', 'year', 'variable']

        validate = validate_params(request.query_params, supported_params)

        if validate:
            return Response(validate, status=400)  # Return a 400 Bad Request response
        
        # query params 
        station = request.query_params.get('station', None)
        year = request.query_params.get('year', latest_year[0] if len(latest_year) else None)
        variable = request.query_params.get('variable', 'pressure')

        if station is not None:
            queryset = queryset.filter(station = station)

        # Extract the month from the received_date and annotate the queryset
        queryset = queryset.filter(variable=variable, received_date__year=year).annotate(
            month=ExtractMonth('received_date')
        )

        # Aggregate the queryset to calculate the average received_rate for each month
        monthly_averages = queryset.values('month').order_by('received_date__month').annotate(
            avg_received_rate=Avg('received_rate'),
            avg_received=Avg('received'),
            avg_expected=Avg('expected'),
        )

        # Map month numbers to month names
        MONTH_NAMES = {
            1: 'January', 2: 'February', 3: 'March', 4: 'April',
            5: 'May', 6: 'June', 7: 'July', 8: 'August',
            9: 'September', 10: 'October', 11: 'November', 12: 'December'
        }
        # Format the result
        result = [
            {
                'month': MONTH_NAMES[month],
                'avg_received_rate': round(avg_rate, 0),  # Round to 0 decimal places
                'avg_received': round(avg_received, 0),  # Round to 0 decimal places
                'avg_expected': round(avg_expected, 0)  # Round to 0 decimal places
            }
            for month, avg_rate, avg_received, avg_expected in monthly_averages.values_list('month', 'avg_received_rate','avg_received', 'avg_expected')
        ]

        return Response(result)


class YearlyTransmissionView(APIView):

    def get(self, request):

        queryset = Transmission.objects.all()
        result = []

        supported_params = ['station', 'variable']

        validate = validate_params(request.query_params, supported_params)

        if validate:
            return Response(validate, status=400)  # Return a 400 Bad Request response

        # query params 
        station = request.query_params.get('station', None)
        variable = request.query_params.get('variable', 'pressure')
        

        if station is not None:
            queryset = queryset.filter(station = station)

        queryset = queryset.filter(variable=variable)

        # Extract the month from the received_date and annotate the queryset
        queryset = queryset.annotate(
            year=ExtractYear('received_date')
        )

        # Aggregate the queryset to calculate the average received_rate for each month
        yearly_averages = queryset.values('year').order_by('received_date__year').annotate(
            avg_received_rate=Avg('received_rate'),
            avg_received=Avg('received'),
            avg_expected=Avg('expected'),
        )

        # Format the result
        result = [
            {
                'year': year,
                'avg_received_rate': round(avg_rate, 0),  # Round to 0 decimal places
                'avg_received': round(avg_received, 0), # Round to 0 decimal places
                'avg_expected': round(avg_expected, 0)  # Round to 0 decimal places
            }
            for year, avg_rate, avg_received, avg_expected in yearly_averages.values_list('year', 'avg_received_rate','avg_received', 'avg_expected')
        ]

        return Response(result)



class AverageMonthlyReceivedRateGeom(APIView):
    def get(self, request):
        month = request.query_params.get('month', None)
        year = request.query_params.get('year', None)
        variable = request.query_params.get('variable', None)

        # Aggregate data by month and calculate average received rate
        monthly_data = Transmission.objects.filter(received_date__month=month,
            received_date__year=year, variable=variable
        ).annotate(
            month=TruncMonth('received_date')
        ).values(
            'month','station__name', 'variable', 'station__wigos_id'
        ).annotate(
            average_received_rate=Avg('received_rate'),
            station_geometry=F('station__geom'),  # Access the geometry field

        )

        # Convert queryset to list of dictionaries
        monthly_data_list = list(monthly_data)

        # Create GeoJSON-like structure
        feature_collection = {
            "type": "FeatureCollection",
            "features": []
        }

        for data in monthly_data_list:
            # Convert Decimal to float for JSON serialization
            average_received_rate = float(data['average_received_rate']) if data['average_received_rate'] is not None else None

            geometry = json.loads(data['station_geometry'].geojson)

            feature = {
                "type": "Feature",
                "properties": {
                    "name": data['station__name'],
                    "wigos_id": data['station__wigos_id'],
                    "month": data['month'].strftime('%Y-%m'),
                    "variable": data['variable'],
                    "average_received_rate": average_received_rate
                },
                "geometry": geometry
            }
            feature_collection["features"].append(feature)

        return Response(feature_collection)



from django.urls import path

from .views import (
    StationListView,
    SynopTransmissionView,
    MonthlyTransmissionView,
    YearlyTransmissionView,
    AverageMonthlyReceivedRateGeom
)

urlpatterns = [
    path('api/synop-transmission-rate/', SynopTransmissionView.as_view(), name='synop-transmission-rate'),
    path('api/monthly-transmission-rate/', MonthlyTransmissionView.as_view(), name='monthly-transmission-rate'),
    path('api/yearly-transmission-rate/', YearlyTransmissionView.as_view(), name='yearly-transmission-rate'),
    path('api/monthly-geom-transmission-rate/', AverageMonthlyReceivedRateGeom.as_view(), name='monthly-geom-transmission-rate'),
    path('api/stations/', StationListView.as_view(), name='station-list'),
]


import logging
import csv
import os
import re
from datetime import datetime, timedelta
from django.contrib.gis.geos import Point
import pandas as pd
import numpy as np

import requests
from django.core.management.base import BaseCommand
from climweb_wdqms.models import Station, Transmission
from adminboundarymanager.models import Country

logger = logging.getLogger(__name__)

# Define the base URL for the WDQMS csv download
BASE_URL = "https://wdqms.wmo.int/wdqmsapi/v1/download/synop/six_hour/availability"

# params---> "date=2024-05-01&period=18&variable=pressure&centers=DWD,ECMWF,JMA,NCEP&baseline=OSCAR"

def download_transmission_rate_csv(date, period, variable, centers, baseline, country_code):

    params = {
        'date': date,
        'period': period,
        'variable': variable,
        'centers': ','.join(centers),
        'baseline': baseline
    }
    file_name = os.path.join("tmp",f"{date}_{period}_{variable}.csv")

    print(f"DOWNLOAD: Starting download of {file_name}")


    response = requests.get(BASE_URL, params=params)

    # Check if the request was successful
    if response.status_code == 200:

        # Check if the directory already exists
        if not os.path.exists("tmp"):
            # Create the directory
            os.makedirs("tmp")

        with open(file_name, 'wb') as f:
            f.write(response.content)

        print(f"DOWNLOAD: {file_name} downloaded successfully.")
        
        # Load the CSV data into a DataFrame
        df = pd.read_csv(file_name)
       
        df_filtered = df[df['country code'] == country_code]

        # Assuming df_filtered is your DataFrame
        df_filtered = df_filtered.copy()
        df_filtered['received_rate'] = (df_filtered['#received'] / df_filtered['#expected']) * 100
         # Group by 'name' and select the row with the highest 'received rate'
        max_rate_indices = df_filtered.groupby('wigosid')['received_rate'].idxmax()
        df_filtered = df_filtered.loc[max_rate_indices]
        df_filtered.replace([np.inf, -np.inf], 0, inplace=True)

        os.remove(file_name)
        return df_filtered

    else:
        print("Failed to retrieve data. Status code:", response.status_code)

def generate_date_range(start_date, end_date):
    dates = []
    current_date = datetime.strptime(start_date, "%Y-%m-%d")
    while current_date <= datetime.strptime(end_date, "%Y-%m-%d"):
        dates.append(current_date.strftime('%Y-%m-%d'))
        current_date += timedelta(days=1)
    return dates


def ingest_transmission_rates(start_date, end_date,variables, periods, centers, country_code):
    dates = generate_date_range(start_date, end_date)
    baseline = "OSCAR" 

    for variable in variables:
        print(f"INGEST: Ingesting {variable.upper()}...")
        variable = variable.lower()
        for date in dates:
            for period in periods:
                trans_rates = download_transmission_rate_csv(date, period, variable, centers, baseline, country_code)

                print(f"INGEST: Starting data ingestion for {date}-{period}")

                # Create or update stations
                stations_to_create = []
                for _, row in trans_rates.iterrows():
                    station_data = {
                        'wigos_id': row['wigosid'],
                        'name': row['name'],
                        'geom':Point(row['longitude'], row['latitude']),
                        'in_oscar':row['in OSCAR']
                    }
                    
                    stations_to_create.append(Station(**station_data))


                # Bulk create new stations
                Station.objects.bulk_create(stations_to_create, ignore_conflicts=True)

                # Create or update transmissions
                transmissions_to_create = []
                for _, row in trans_rates.iterrows():

                    if Transmission.objects.filter(station=row['wigosid'], variable=row['variable'], received_date=datetime.strptime(row['date'],'%Y-%m-%d %H:%M:%S%z')).exists():
                        transmission_data = {
                            'received_rate':row['received_rate'],
                            'received':row['#received'],
                            'expected':row['#expected'],
                        }
                        Transmission.objects.filter(station=row['wigosid'], variable=row['variable'], received_date=datetime.strptime(row['date'],'%Y-%m-%d %H:%M:%S%z')).update(**transmission_data)

                    else:
                        station = Station.objects.get(wigos_id=row['wigosid'])
                        transmissions_to_create.append(Transmission(
                            station=station,
                            variable=row['variable'],
                            received_rate=row['received_rate'],
                            received=row['#received'],
                            expected=row['#expected'],
                            received_date=datetime.strptime(row['date'],'%Y-%m-%d %H:%M:%S%z')
                        ))

                # Bulk create new transmissions
                Transmission.objects.bulk_create(transmissions_to_create, ignore_conflicts=True)
                
                print(f"INGEST: Completed ingestion for {date}-{period}")



class Command(BaseCommand):
    help = ('Fetch Country level transmission rate from WDQMS')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def add_arguments(self, parser):
        parser.add_argument('-s', '--start_date', type=str, help='Start date of transmission. format YYYY-MM-DD') 
        parser.add_argument('-e', '--end_date', type=str, help='End date of transmission. format YYYY-MM-DD') 
        parser.add_argument('-var', '--variables', nargs='+', type=str, help='List of variables e.g pressure,temperature, humidity, meridional_wind, zonal_wind') 
        parser.add_argument('-p', '--periods', nargs='+', type=str, help='List of synoptic hours e.g 00, 06, 12, 18') 
        parser.add_argument('-c', '--centers', nargs='+', type=str, help='List of monitoring centers e.g DWD, ECMWF, JMA, NCEP') 

        # Arguments are not added here since they will be parsed manually
        return

    def handle(self, *args, **kwargs):

        latest_date = Transmission.objects.values_list('received_date__date').order_by('received_date__date').last()
        yesterday = datetime.now().date() - timedelta(days=1)

        if kwargs['start_date'] is not None:
            # check first cli params
            start_date = kwargs['start_date']
        elif latest_date:
            # check latest date in db 
            start_date = latest_date[0].strftime("%Y-%m-%d")
        else:
            # use earliest date in wdqms. usually means the db is empty 
            start_date = "2019-01-01" 


        end_date = kwargs['end_date'] if kwargs['end_date'] is not None else yesterday.strftime("%Y-%m-%d")
        periods = kwargs['periods'] if kwargs['periods'] is not None else ["00", "06", "12", "18"]
        centers = kwargs['centers'] if kwargs['centers'] is not None else ["DWD", "ECMWF", "JMA", "NCEP"]
        variables = kwargs['variables'] if kwargs['variables'] is not None else ['pressure', 'temperature', 'humidity', 'meridional_wind' , 'zonal_wind']

        variables_ls = ['pressure', 'temperature', 'humidity', 'meridional_wind' , 'zonal_wind']
        period_ls = ["00", "06", "12", "18"]
        center_ls = ["DWD", "ECMWF", "JMA", "NCEP"]

        # Regular expression to match YYYY-MM-DD format
        date_pattern = re.compile(r'^\d{4}-\d{2}-\d{2}$')

        # Parsing the arguments manually
        if start_date is not None:
            if not date_pattern.match(start_date):
                self.stderr.write(self.style.ERROR(f"Invalid format for 'start_date'. Use YYYY-MM-DD format."))
                return  # Exit the command

        if end_date is not None:  
            if not date_pattern.match(end_date):
                self.stderr.write(self.style.ERROR(f"Invalid format for 'end_date'. Use YYYY-MM-DD format."))
                return  # Exit the command
            
        if datetime.strptime(start_date, "%Y-%m-%d") > datetime.strptime(end_date, "%Y-%m-%d"):
            self.stderr.write(self.style.ERROR(f"'End date' cannot come earlier than 'Start date'"))
            return  # Exit the command


        if variables is not None:
            for variable in variables:
                if variable not in variables_ls:
                    self.stderr.write(self.style.ERROR(f"Allowed variables include pressure,temperature, humidity, meridional_wind, zonal_wind"))
                    return  # Exit the command

            # Split the value by comma and append to list_arg
        if periods is not None:
            for period in periods:
                if period not in period_ls:
                    self.stderr.write(self.style.ERROR(f"'{period}' is not a valid option. Choices are 00 06 12 18"))
                    return  # Exit the command
            
        if centers is not None:
            for center in centers:
                if center.upper() not in center_ls:
                    self.stderr.write(self.style.ERROR(f"'{center}' is not a valid option. Choices are DWD ECMWF JMA NCEP"))
                    return  # Exit the command
                
            

        if start_date is not None and end_date is not None and variables is not None and centers is not None and periods is not None:
            for country in Country.objects.all():
                if Country.objects.count() > 0:
                    # loop through all countries in boundary manager for data fetching and ingestion
                    self.stdout.write(f"FETCH: Requesting data for {country.country.name}")

                    ingest_transmission_rates(start_date, end_date, variables, periods, centers, country.country.alpha3)
                else:
                    self.stderr.write(self.style.ERROR(f"Please select atleast one country in admin boundary settings first"))


        


            

        
        




        


from rest_framework import serializers

from climweb_wdqms.models import Transmission, Station


class StationSerializer(serializers.ModelSerializer):
    class Meta:
        model = Station
        # fields = ('wigos_id', 'name', "in_oscar", "geom")
        fields = "__all__"

    
    def to_representation(self, instance):
        representation = super().to_representation(instance)
        feature = {
            'type': 'Feature',
            'geometry': {
                'type': 'Point',
                'coordinates': [instance.geom.x, instance.geom.y]
            },
            'properties': {}
        }
        # Set all other fields as properties
        for field in representation:
            if field != 'geom':
                feature['properties'][field] = representation[field]
        return feature
        return representation

class TransmissionSerializer(serializers.ModelSerializer):
    class Meta:
        model = Transmission
        fields = ["received_date", "station", "variable", "received_rate", "recieved", "expected"]




from django.contrib import admin

from .models import Station, Transmission


# Register your models here.

class TransmissionModelAdmin(admin.ModelAdmin):
    list_filter = ('variable', 'received_date')
    search_fields = [ 'station__name', 'received_date']
    list_display = ('station', 'received_date', 'variable', 'received_rate')



admin.site.register(Station)
admin.site.register(Transmission,TransmissionModelAdmin)





# Generated by Django 4.2.11 on 2024-05-12 00:08

import django.contrib.gis.db.models.fields
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Station',
            fields=[
                ('wigos_id', models.CharField(max_length=50, primary_key=True, serialize=False, verbose_name='Wigos ID')),
                ('name', models.CharField(max_length=255, verbose_name='Station name')),
                ('geom', django.contrib.gis.db.models.fields.PointField(srid=4326, verbose_name='Geometry')),
                ('in_oscar', models.BooleanField(verbose_name='In Oscar?')),
            ],
            options={
                'verbose_name': 'Station',
                'verbose_name_plural': 'Stations',
            },
        ),
        migrations.CreateModel(
            name='Transmission',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('variable', models.CharField(max_length=50, verbose_name='Transmission Variable')),
                ('received', models.IntegerField(null=True, verbose_name='Transmissions received')),
                ('expected', models.IntegerField(null=True, verbose_name='Transmissions expected')),
                ('received_rate', models.DecimalField(decimal_places=2, max_digits=5, verbose_name='Transmission Rate')),
                ('received_date', models.DateTimeField(verbose_name='Date Time Received')),
                ('station', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='climweb_wdqms.station')),
            ],
            options={
                'verbose_name': 'Transmission',
                'verbose_name_plural': 'Transmissions',
            },
        ),
    ]
